#version 120

// Enhanced Minecraft Shader for Iris
// Created: March 2025

/* ------ Vertex Shader ------ */
#ifdef VERTEX

varying vec4 texcoord;
varying vec4 color;
varying vec4 lmcoord;
varying vec3 normal;
varying vec3 worldPosition;
varying float viewDistance;
varying vec3 tangent;
varying vec3 binormal;
varying vec3 viewVector;

uniform mat4 gbufferModelView;
uniform mat4 gbufferModelViewInverse;
uniform vec3 shadowLightPosition;
uniform float frameTimeCounter;

attribute vec4 at_tangent;

void main() {
    // Pass texture coordinates to fragment shader
    texcoord = gl_TextureMatrix[0] * gl_MultiTexCoord0;
    
    // Pass vertex color to fragment shader
    color = gl_Color;
    
    // Pass lightmap coordinates to fragment shader
    lmcoord = gl_TextureMatrix[1] * gl_MultiTexCoord1;
    
    // Transform normal to world space
    normal = normalize(gl_NormalMatrix * gl_Normal);
    
    // Calculate tangent and binormal for normal mapping
    tangent = normalize(gl_NormalMatrix * at_tangent.xyz);
    binormal = cross(normal, tangent) * at_tangent.w;
    
    // Calculate world position
    vec4 position = gbufferModelViewInverse * gl_ModelViewMatrix * gl_Vertex;
    worldPosition = position.xyz;
    
    // Calculate view distance for fog
    viewDistance = length((gbufferModelView * position).xyz);
    
    // Calculate view vector for reflections
    viewVector = normalize(gbufferModelViewInverse[3].xyz - worldPosition);
    
    // Apply subtle wind animation to foliage
    if (gl_Color.g > 0.5 && gl_Color.r < 0.1) { // Detect foliage by color
        float windStrength = 0.05;
        float windSpeed = 1.5;
        float windFactor = sin(frameTimeCounter * windSpeed + worldPosition.x * 0.1 + worldPosition.z * 0.1) * 
                          cos(frameTimeCounter * windSpeed * 0.7 + worldPosition.x * 0.1) * windStrength;
        
        // Apply wind only to the top of plants
        float heightFactor = gl_Vertex.y - floor(gl_Vertex.y);
        gl_Vertex.x += windFactor * heightFactor;
        gl_Vertex.z += windFactor * heightFactor * 0.5;
    }
    
    // Set final vertex position
    gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex;
}
#endif

/* ------ Fragment Shader ------ */
#ifdef FRAGMENT

varying vec4 texcoord;
varying vec4 color;
varying vec4 lmcoord;
varying vec3 normal;
varying vec3 worldPosition;
varying float viewDistance;
varying vec3 tangent;
varying vec3 binormal;
varying vec3 viewVector;

uniform sampler2D texture;
uniform sampler2D lightmap;
uniform float rainStrength;
uniform int worldTime;
uniform vec3 fogColor;
uniform vec3 skyColor;
uniform float fogStart;
uniform float fogEnd;
uniform vec3 shadowLightPosition;
uniform float frameTimeCounter;
uniform float viewWidth;
uniform float viewHeight;

// Custom adjustable parameters
const float saturationAdjust = 1.3;        // Increase color saturation
const float brightnessAdjust = 1.2;        // Increase brightness
const float contrastAdjust = 1.1;          // Increase contrast
const float sunlightIntensity = 1.7;       // Increase sunlight brightness
const float ambientLightBoost = 1.4;       // Boost ambient lighting
const float waterWaveSpeed = 0.04;         // Water wave animation speed
const float waterWaveStrength = 0.03;      // Water wave height
const float waterTransparency = 0.85;      // Water transparency
const float reflectionStrength = 0.6;      // Reflection strength on water
const float shadowStrength = 0.7;          // Strength of soft shadows
const float bloomStrength = 0.25;          // Strength of bloom effect
const float sharpnessAmount = 0.3;         // Image sharpening amount
const float dofStrength = 0.4;             // Depth of field strength
const float dofFocusDistance = 20.0;       // Distance at which things are in focus
const float fogDensity = 0.8;              // Fog density
const float nightBrightness = 1.5;         // Night brightness boost
const float wettingSaturation = 0.7;       // How much rain reduces saturation

// Toggle Features
const bool enableImprovedWater = true;     // Water effects
const bool enableWindEffect = true;        // Wind animation on foliage
const bool enableBloom = true;             // Bloom/glow effect
const bool enableSharpening = true;        // Image sharpening
const bool enableDepthOfField = true;      // Depth of field
const bool enableReflections = true;       // Reflections on water surfaces
const bool enableGodrays = true;           // Light shafts/god rays
const bool enableDynamicWeather = true;    // Weather-based effects
const bool enableTonemapping = true;       // HDR-like tone mapping
const bool enableVignette = true;          // Screen darkening at edges
const bool enableWaving = true;            // Waving grass/leaves

// Converts RGB to HSV color space
vec3 rgb2hsv(vec3 c) {
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));
    
    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}

// Converts HSV to RGB color space
vec3 hsv2rgb(vec3 c) {
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

// ACES tone mapping function
vec3 toneMap(vec3 color) {
    float a = 2.51;
    float b = 0.03;
    float c = 2.43;
    float d = 0.59;
    float e = 0.14;
    return clamp((color * (a * color + b)) / (color * (c * color + d) + e), 0.0, 1.0);
}

// Calculates a simple pseudo-random value based on coordinates
float hash(vec2 p) {
    return fract(sin(dot(p, vec2(12.9898, 78.233))) * 43758.5453);
}

// Apply color adjustments including day/night adaptation
vec3 adjustColor(vec3 color) {
    // Determine if it's night or day based on sky color
    bool isNight = skyColor.r < 0.3 && skyColor.g < 0.3 && skyColor.b < 0.3;
    
    // Convert to HSV for easier adjustment
    vec3 hsv = rgb2hsv(color);
    
    // Weather adjustments
    float rainFactor = rainStrength * 0.7;
    
    // During rain, reduce saturation but maintain brightness
    if (enableDynamicWeather && rainStrength > 0.1) {
        hsv.y *= mix(1.0, wettingSaturation, rainFactor);
        // Make colors slightly cooler during rain
        hsv.x = mix(hsv.x, 0.6, rainFactor * 0.1); // Shift toward blue
    }
    
    // Night-specific adjustments
    if (isNight) {
        // Boost brightness at night
        hsv.z = pow(hsv.z, 0.7) * nightBrightness;
        // Shift colors slightly toward blue-purple at night
        hsv.x = mix(hsv.x, 0.7, 0.1);
        // Reduce saturation slightly at night
        hsv.y *= 0.8;
    } else {
        // Day adjustments - more saturation
        hsv.y *= saturationAdjust;
        // Normal brightness adjustment
        hsv.z *= brightnessAdjust;
    }
    
    // Convert back to RGB
    vec3 rgb = hsv2rgb(hsv);
    
    // Apply contrast (different values for day/night)
    float contrastValue = isNight ? contrastAdjust * 0.8 : contrastAdjust;
    rgb = (rgb - 0.5) * contrastValue + 0.5;
    
    return clamp(rgb, 0.0, 1.0);
}

// Depth of field effect
vec3 applyDOF(vec3 color, float depth) {
    if (!enableDepthOfField) return color;
    
    // Calculate blur amount based on distance from focus point
    float focusPoint = dofFocusDistance;
    float blurFactor = abs(depth - focusPoint) / 50.0 * dofStrength;
    blurFactor = clamp(blurFactor, 0.0, 1.0);
    
    // Basic box blur based on blur factor
    vec3 blurredColor = vec3(0.0);
    float kernelSize = 3.0 * blurFactor;
    float samples = 0.0;
    
    for (float x = -kernelSize; x <= kernelSize; x += 1.0) {
        for (float y = -kernelSize; y <= kernelSize; y += 1.0) {
            vec2 offset = vec2(x, y) / vec2(viewWidth, viewHeight);
            blurredColor += texture2D(texture, texcoord.st + offset).rgb;
            samples += 1.0;
        }
    }
    
    blurredColor /= samples;
    
    // Blend between sharp and blurred based on blur factor
    return mix(color, blurredColor, blurFactor);
}

// Apply image sharpening
vec3 applySharpening(sampler2D tex, vec2 uv) {
    if (!enableSharpening) return texture2D(tex, uv).rgb;
    
    vec2 texSize = vec2(viewWidth, viewHeight);
    vec2 texelSize = 1.0 / texSize;
    
    // Sample neighboring pixels
    vec3 center = texture2D(tex, uv).rgb;
    vec3 top = texture2D(tex, uv + vec2(0, -1) * texelSize).rgb;
    vec3 right = texture2D(tex, uv + vec2(1, 0) * texelSize).rgb;
    vec3 bottom = texture2D(tex, uv + vec2(0, 1) * texelSize).rgb;
    vec3 left = texture2D(tex, uv + vec2(-1, 0) * texelSize).rgb;
    
    // Apply sharpening filter
    vec3 sharpenedColor = center * (1.0 + 4.0 * sharpnessAmount) - (top + right + bottom + left) * sharpnessAmount;
    
    return clamp(sharpenedColor, 0.0, 1.0);
}

// Bloom/glow effect
vec3 applyBloom(vec3 color) {
    if (!enableBloom) return color;
    
    // Extract bright parts of the image
    vec3 brightPass = max(color - 0.7, 0.0) * 2.0;
    
    // Blur the bright parts (simple approximation)
    vec3 blur = vec3(0.0);
    float totalWeight = 0.0;
    
    for (float x = -3.0; x <= 3.0; x += 1.0) {
        for (float y = -3.0; y <= 3.0; y += 1.0) {
            float weight = 1.0 / (1.0 + x*x + y*y);
            vec2 offset = vec2(x, y) / vec2(viewWidth, viewHeight) * 2.0;
            blur += texture2D(texture, texcoord.st + offset).rgb * weight;
            totalWeight += weight;
        }
    }
    
    blur /= totalWeight;
    blur *= brightPass;
    
    // Add bloom to original color
    return color + blur * bloomStrength;
}

// Light shaft/god ray effect
vec3 applyGodrays(vec3 color) {
    if (!enableGodrays) return color;
    
    // Get sun direction (simplified)
    vec3 sunDir = normalize(shadowLightPosition);
    
    // Raymarch towards sun (simplified version)
    vec3 rayDir = normalize(vec3(texcoord.s - 0.5, texcoord.t - 0.5, 1.0));
    float alignment = max(0.0, dot(rayDir, sunDir));
    
    // Ray strength based on sun alignment and weather
    float rayStrength = pow(alignment, 6.0) * (1.0 - rainStrength * 0.8);
    
    // Add rays to the original color
    vec3 rayColor = mix(skyColor, vec3(1.0, 0.9, 0.7), 0.5);
    return color + rayColor * rayStrength * 0.2;
}

// Vignette effect
vec3 applyVignette(vec3 color, vec2 uv) {
    if (!enableVignette) return color;
    
    // Calculate distance from center
    vec2 center = vec2(0.5, 0.5);
    float dist = length(uv - center) * 1.5;
    
    // Apply vignette
    float vignette = smoothstep(0.8, 0.2, dist);
    return color * vignette;
}

// Water animation and effects
vec4 handleWater(vec4 baseColor, vec2 texCoords) {
    if (!enableImprovedWater) return baseColor;
    
    // Check if this is water (approximate by using base color)
    bool isWater = baseColor.b > 0.6 && baseColor.g > 0.4 && baseColor.r < 0.5;
    
    if (isWater) {
        // Calculate wave effect
        float time = frameTimeCounter * waterWaveSpeed;
        vec2 waveOffset = vec2(
            sin(texCoords.x * 10.0 + time + texCoords.y * 5.0) * waterWaveStrength,
            cos(texCoords.y * 10.0 + time * 0.7 + texCoords.x * 6.0) * waterWaveStrength
        );
        
        // Multi-layered waves for more realism
        waveOffset += vec2(
            sin(texCoords.x * 20.0 - time * 0.8) * waterWaveStrength * 0.5,
            cos(texCoords.y * 15.0 - time * 0.9) * waterWaveStrength * 0.5
        );
        
        // Apply wave distortion to texture coordinates
        vec2 distortedCoords = texCoords + waveOffset;
        vec4 distortedColor = texture2D(texture, distortedCoords);
        
        // Blend based on distance from original
        float blendFactor = length(waveOffset) * 20.0;
        baseColor = mix(baseColor, distortedColor, clamp(blendFactor, 0.0, 0.5));
        
        // Add a slight blue tint and transparency
        baseColor.rgb = mix(baseColor.rgb, vec3(0.2, 0.5, 0.9), 0.2);
        baseColor.a = waterTransparency;
        
        // Add reflections if enabled
        if (enableReflections) {
            // Calculate reflection vector
            vec3 reflectDir = reflect(-viewVector, normal);
            
            // Sample environment for reflection (simplified - just use sky color)
            vec3 reflectionColor = mix(skyColor, vec3(1.0), 0.5);
            
            // Apply Fresnel effect - more reflective at glancing angles
            float fresnel = pow(1.0 - dot(normal, -viewVector), 5.0);
            
            // Add reflection to water
            baseColor.rgb = mix(baseColor.rgb, reflectionColor, fresnel * reflectionStrength);
            
            // Add sparkles to water
            float sparkle = pow(max(0.0, dot(normal, normalize(shadowLightPosition))), 50.0);
            baseColor.rgb += vec3(sparkle) * (1.0 - rainStrength) * 0.3;
        }
    }
    
    return baseColor;
}

// Apply atmospheric fog with improved aesthetics
vec3 applyFog(vec3 color, float distance) {
    // Early return if distance is very small
    if (distance < fogStart) return color;
    
    // Adjust fog based on weather and time
    bool isNight = skyColor.r < 0.3 && skyColor.g < 0.3 && skyColor.b < 0.3;
    
    // Calculate fog color based on time and weather
    vec3 adjustedFogColor;
    
    if (rainStrength > 0.1) {
        // Rainy fog is grayer and thicker
        adjustedFogColor = mix(fogColor, vec3(0.5, 0.5, 0.5), rainStrength * 0.7);
    } else if (isNight) {
        // Night fog is darker and slightly blue-tinted
        adjustedFogColor = mix(fogColor, vec3(0.05, 0.05, 0.1), 0.6);
    } else {
        // Day fog uses sky color with slight adjustments
        adjustedFogColor = mix(fogColor, skyColor, 0.7);
    }
    
    // Calculate fog factor with exponential falloff for more realism
    float fogFactor;
    
    if (rainStrength > 0.1) {
        // Denser fog during rain
        fogFactor = 1.0 - exp(-distance * 0.02 * fogDensity * (1.0 + rainStrength));
    } else {
        // Normal fog calculation
        fogFactor = 1.0 - exp(-distance * 0.01 * fogDensity);
    }
    
    fogFactor = clamp(fogFactor, 0.0, 0.95); // Cap maximum fog
    
    // Height-based fog reduction (less fog at higher elevations)
    float heightFactor = clamp(worldPosition.y / 100.0, 0.0, 0.3);
    fogFactor *= (1.0 - heightFactor);
    
    // Apply fog blend
    return mix(color, adjustedFogColor, fogFactor);
}

void main() {
    // Get base color from texture with sharpening if enabled
    vec3 baseColorRGB = enableSharpening ? 
        applySharpening(texture, texcoord.st) : 
        texture2D(texture, texcoord.st).rgb;
    
    vec4 baseColor = vec4(baseColorRGB, texture2D(texture, texcoord.st).a);
    
    // Apply water effects if needed
    baseColor = handleWater(baseColor, texcoord.st);
    
    // Get lighting from lightmap
    vec3 lighting = texture2D(lightmap, lmcoord.st).rgb;
    
    // Dynamic day/night lighting
    bool isNight = skyColor.r < 0.3 && skyColor.g < 0.3 && skyColor.b < 0.3;
    float timeAdjust = isNight ? nightBrightness : 1.0;
    
    // Boost sunlight and ambient light
    vec3 sunlight = vec3(lighting.r * sunlightIntensity * timeAdjust);
    vec3 ambientLight = vec3(lighting.b * ambientLightBoost * timeAdjust);
    lighting = max(sunlight, ambientLight);
    
    // Apply vertex color
    vec3 finalColor = baseColor.rgb * color.rgb;
    
    // Apply lighting
    finalColor *= lighting;
    
    // Apply color adjustments
    finalColor = adjustColor(finalColor);
    
    // Apply bloom effect
    finalColor = applyBloom(finalColor);
    
    // Apply godrays/light shafts
    finalColor = applyGodrays(finalColor);
    
    // Apply depth of field effect
    finalColor = applyDOF(finalColor, viewDistance);
    
    // Apply fog
    finalColor = applyFog(finalColor, viewDistance);
    
    // Apply vignette effect
    finalColor = applyVignette(finalColor, texcoord.st);
    
    // Apply tone mapping
    if (enableTonemapping) {
        finalColor = toneMap(finalColor);
    }
    
    // Set final output color
    gl_FragColor = vec4(finalColor, baseColor.a * color.a);
}
#endif